from typing import Any, Dict, List, Optional, Tuple, Type, TypeVar, Union
from typing_extensions import TypeAlias

import attr


# @attr.s(auto_attribs=True)
class PynalyserType:
    @property
    def as_str(self) -> str:
        raise NotImplementedError

    def deref(self) -> "PynalyserType":
        return self


@attr.s(auto_attribs=True, hash=True)
class UnionType(PynalyserType):
    types: Tuple[PynalyserType, ...] = attr.ib()

    @types.validator  # type: ignore
    def check(self, attribute: Any, value: Tuple[PynalyserType, ...]):
        if len(value) <= 1:
            raise ValueError("there should be more than 2 different types")

    @property
    def as_str(self) -> str:
        return f"Union[{', '.join(tp.as_str for tp in self.types)}]"

    @classmethod
    def make(cls, *types: PynalyserType,
             fallback: Optional[PynalyserType] = None) -> PynalyserType:
        new_types: List[PynalyserType] = []

        for tp in types:
            if isinstance(tp, cls):
                new_types.extend(tp.types)
            else:
                new_types.append(tp)

        unique_types = tuple(set(new_types))
        if len(unique_types) == 0:
            if fallback is not None:
                return fallback
            raise ValueError("length of types should not be 0")
        elif len(unique_types) == 1:
            return types[0]
        else:
            return cls(unique_types)


DUNDER_SIGNATURE = Tuple[Type["SingleType"], ...]


@attr.s(auto_attribs=True, hash=True)
class SingleType(PynalyserType):
    name: str = attr.ib(kw_only=True)
    is_builtin: bool = attr.ib(kw_only=True)

    dunder_signatures: Dict[str, DUNDER_SIGNATURE] = attr.ib(
        factory=dict, init=False, hash=False)

    is_completed: bool = attr.ib(default=False, kw_only=True)

    @property
    def as_str(self) -> str:
        return self.name


AnyType = SingleType(name="object", is_builtin=False)
UnknownType = SingleType(name="object", is_builtin=False)


# T = TypeVar("T")

NotImplementedType = type(NotImplemented)
ReturnT = Union[PynalyserType, NotImplementedType]  # type: ignore


@attr.s(auto_attribs=True, hash=True)
class IntType(SingleType):
    name: str = "int"
    is_builtin: bool = True

    dunder_signatures: Dict[str, DUNDER_SIGNATURE] = attr.ib(
        factory=lambda: {
            "lt": (IntType,),
            "gt": (IntType,),
            "le": (IntType,),
            "ge": (IntType,),
            "eq": (IntType,),
            "ne": (IntType,),
        },
        init=False, hash=False)

    is_completed: bool = True

    # also see "long_compare" in cpython github
    def _cmp(self, other: PynalyserType) -> ReturnT:
        if isinstance(other, IntType):
            return BoolType()
        return NotImplemented

    __lt__ = _cmp  # type: ignore
    __gt__ = _cmp  # type: ignore
    __le__ = _cmp  # type: ignore
    __ge__ = _cmp  # type: ignore
    __eq__ = _cmp  # type: ignore
    __ne__ = _cmp  # type: ignore

    # dunder_signatures["lt"] = [IntType]
    # dunder_signatures["gt"] = [IntType]
    # dunder_signatures["le"] = [IntType]
    # dunder_signatures["ge"] = [IntType]
    # dunder_signatures["eq"] = [IntType]
    # dunder_signatures["ne"] = [IntType]


@attr.s(auto_attribs=True, hash=True)
class BoolType(IntType):
    name: str = "bool"
    is_builtin: bool = True

    is_completed: bool = True


@attr.s(auto_attribs=True)
class IterableType(SingleType):
    item_type: PynalyserType
    name: str = "Iterable"

    @property
    def as_str(self) -> str:
        return f"{super().as_str}[{self.item_type.as_str}]"


@attr.s(auto_attribs=True, hash=True, auto_detect=True)
class SequenceType(IterableType):
    # TODO: not finshed, see docs.python.org/3/library/collections.abc.html

    dunder_signatures: Dict[str, DUNDER_SIGNATURE] = attr.ib(
        factory=lambda: {
            "mul":     (IntType,      ),
            "rmul":    (IntType,      ),
            "getitem": (type(AnyType),)
        },
        init=False, hash=False)

    # see docs.python.org/3/c-api/typeobj.html
    def __mul__(self, other: PynalyserType) -> ReturnT:
        if isinstance(other, IntType):
            return self
        return NotImplemented
    # dunder_signatures["mul"] = [IntType]

    __rmul__ = __mul__
    # dunder_signatures["rmul"] = [IntType]

    def __getitem__(self, item: PynalyserType) -> PynalyserType:
        return AnyType


@attr.s(auto_attribs=True, hash=True)
class SliceType(SingleType):
    name: str = "slice"
    is_builtin: bool = True

    is_completed: bool = True


@attr.s(auto_attribs=True, hash=True)
class ListType(SequenceType):
    name: str = "list"
    is_builtin: bool = True

    dunder_signatures: Dict[str, DUNDER_SIGNATURE] = attr.ib(
        factory=lambda: {
            "mul":     (IntType,           ),
            "rmul":    (IntType,           ),
            "getitem": (IntType, SliceType,),
        },
        init=False, hash=False)

    is_completed: bool = True

    def __getitem__(self, item: PynalyserType) -> PynalyserType:
        if isinstance(item, IntType):
            return self.item_type
        if isinstance(item, SliceType):
            return self
        return NotImplemented


@attr.s(auto_attribs=True, hash=True)
class TupleType(SequenceType):
    name: str = "tuple"
    is_builtin: bool = True

    is_completed: bool = True



import sys
import typing
from ast import *

if sys.version_info < (3, 10):
    class pattern(AST):
        ...

    class match_case(AST):
        pattern: pattern
        guard: typing.Optional[expr]
        body: typing.List[stmt]

    class Match(stmt):
        subject: expr
        cases: typing.List[match_case]

if sys.version_info < (3, 8):
    class NamedExpr(expr):
        target: expr
        value: Expr


# TODO: Create dummy classes for ast.Num, ast.Str, ast.Bytes,
# ast.NameConstant, ast.Ellipsis, ast.Index and ast.ExtSlice
# for the version where they will be removed
# and add them to ast
# (For the purposes of correct type checking and
# maintaining consistent content of the module ast)

# TODO: build python parser from for the newest version of cpython
# and ship it in a separate package to be python version independent
# or use other python parsers

del sys, typing



from typing import List, Optional, Tuple

import attr

from ._types import (AnyType, IntType, IterableType, PynalyserType, SingleType,
                     UnionType, UnknownType)
from .symbols import SymbolData


@attr.s(auto_attribs=True, auto_detect=True)
class SymbolType(PynalyserType):
    name: str
    symbol: SymbolData

    def deref(self) -> PynalyserType:
        return self.symbol.type.deref()

    def __hash__(self) -> int:
        return hash((type(self), self.name))


binop_s = {"Add": "+",
           "Sub": "-",
           "Mult": "*",
           "MatMult": "@",
           "Div": "/",
           "Mod": "%",
           "LShift": "<<",
           "RShift": ">>",
           "BitOr": "|",
           "BitXor": "^",
           "BitAnd": "&",
           "FloorDiv": "//",
           "Pow": "**"}

binop = {
    name: eval(f"lambda a, b: a {op} b")
    for name, op in binop_s.items()}


op_to_dunder = {
    "Mult": "mul"
}


@attr.s(auto_attribs=True, hash=True)
class BinOpType(PynalyserType):
    op: str
    left: PynalyserType
    right: PynalyserType

    def __attrs_post_init__(self):
        left = self.left.deref()
        assert isinstance(left, SingleType)
        right = self.right.deref()
        assert isinstance(right, SingleType)

        def narrow_type(tp: SingleType, both: Tuple[str, str],
                        method: str, use_left: bool):
            if use_left:
                this_method = method
                other_method = "r" + method
            else:
                this_method = "r" + method
                other_method = method

            sign = tp.dunder_signatures.get(this_method, None)
            if sign is not None:
                # TODO: also add type that have __{other_method}__ op
                other = UnionType.make(*(
                    _tp()  # type: ignore
                    for _tp in sign))
                assert isinstance(other, SingleType)
            else:
                print(tp.dunder_signatures, method, sign)
                raise TypeError(
                    f"unsupported operand type(s) for {binop_s[self.op]}: "
                    f"'{both[0]}' and '{both[1]}'")

            return other

        if not left.is_completed and not right.is_completed:
            raise NotImplementedError
        elif not left.is_completed:
            left = narrow_type(
                right, (right.name, left.name),
                op_to_dunder[self.op], use_left=False)
        elif not right.is_completed:
            right = narrow_type(
                left, (left.name, right.name),
                op_to_dunder[self.op], use_left=True)
        # else:
        #     raise NotImplementedError
        #     name = op_to_dunder[op]
        #     lsig = left.dunder_signatures.get(name, None)
        #     if lsig is not None:
        #         if type(right) in lsig:
        #             return getattr(right, name)(left)

        #     name = "r" + name
        #     rsig = right.dunder_signatures.get(name, None)
        #     if rsig is not None:
        #         if type(left) in rsig:
        #             return getattr(left, name)(right)
        #     elif lsig is None:
        #         raise TypeError("")

        if isinstance(self.left, SymbolType):
            self.left.symbol.type = left
        if isinstance(self.right, SymbolType):
            self.right.symbol.type = right

    def deref(self) -> PynalyserType:
        return binop[self.op](self.left.deref(), self.right.deref())


cmp_s = {"Lt": "<"}

cmp = {
    name: eval(f"lambda a, b: a {op} b")
    for name, op in cmp_s.items()}

cmp_to_dunder = {"Lt": "lt"}


@attr.s(auto_attribs=True, hash=True)
class CompareType(PynalyserType):
    left: PynalyserType
    ops: List[str]
    comparators: List[PynalyserType]

    def __attrs_post_init__(self):
        left = self.left.deref()
        assert isinstance(left, SingleType)
        right = self.comparators[0].deref()
        assert isinstance(right, SingleType)

        def narrow_type(tp: SingleType, both: Tuple[str, str],
                        method: str, use_left: bool):
            if use_left:
                this_method = method
                # other_method = "r" + method
                # swapped_cmp_op - lt <=> gt ..
            else:
                this_method = "r" + method
                # other_method = method

            sign = tp.dunder_signatures.get(this_method, None)
            if sign is not None:
                # TODO: also add type that have __{other_method}__ op
                other = UnionType.make(*(
                    _tp()  # type: ignore
                    for _tp in sign))
                assert isinstance(other, SingleType)
            else:
                print(tp.dunder_signatures, method, sign)
                raise TypeError(
                    f"unsupported operand type(s) for {cmp_s[self.ops[0]]}: "
                    f"'{both[0]}' and '{both[1]}'")

            return other

        if not left.is_completed and not right.is_completed:
            pass
            # raise NotImplementedError
        elif not left.is_completed:
            left = narrow_type(
                right, (right.name, left.name),
                cmp_to_dunder[self.ops[0]], use_left=False)
        elif not right.is_completed:
            right = narrow_type(
                left, (left.name, right.name),
                cmp_to_dunder[self.ops[0]], use_left=True)

        if isinstance(self.left, SymbolType):
            self.left.symbol.type = left
        if isinstance(self.comparators[0], SymbolType):
            self.comparators[0].symbol.type = right

    def _deref(self, res: PynalyserType, ops,
               vals: List[PynalyserType]) -> PynalyserType:
        if ops and vals:
            res = ops.pop(0)(res, vals.pop(0))
            if ops and vals:
                return self._deref(res, ops, vals)
            else:
                return res
        raise ValueError("Empty 'ops' and 'val'")

    def deref(self) -> PynalyserType:
        return self._deref(self.left, self.ops, self.comparators)


@attr.s(auto_attribs=True, hash=True)
class SubscriptType(PynalyserType):
    value: PynalyserType
    slice: PynalyserType

    def deref(self) -> PynalyserType:
        return self.value.deref()[self.slice.deref()]  # type: ignore

    # def visit_Subscript(self, node: ast.Subscript) -> PynalyserType:
    #     value_type = self.visit(node.value)
    #     slice_type = self.visit(node.slice)
    #     if isinstance(value_type, SequenceType):
    #         return value_type[slice_type]

    #     raise NotImplementedError


@attr.s(auto_attribs=True, hash=True)
class ItemType(PynalyserType):
    iterable: PynalyserType

    def __attrs_post_init__(self):
        tp = self.iterable.deref()
        if tp is UnknownType:
            if isinstance(self.iterable, SymbolType):
                self.iterable.symbol.type = IterableType(
                    item_type=UnknownType, is_builtin=False)
            # XXX: should we let it be and cause an error?
            # else:
            #     self.iterable = IterableType(
            #         item_type=UnknownType, is_builtin=False)

    def deref(self) -> PynalyserType:
        return self.iterable.deref().item_type.deref()  # type: ignore


@attr.s(auto_attribs=True, hash=True)
class CallType(PynalyserType):
    func: PynalyserType
    args: Tuple[PynalyserType, ...]
    keywords: Tuple[Tuple[Optional[str], PynalyserType], ...]

    def deref(self) -> PynalyserType:
        if isinstance(self.func, SymbolType) and self.func.name == "range":
            return IterableType(item_type=IntType(),
                                is_builtin=False)  # XXX: is_builtin??

        return AnyType



from enum import Flag, auto
from typing import DefaultDict

import attr

from ._types import PynalyserType, UnknownType


class ScopeType(Flag):  # XXX: maybe ScopeT?
    GLOBAL = auto()  # used with global keyword in this scope
    NONLOCAL = auto()  # used with nonlocal keyword in this scope
    LOCAL = auto()  # has an assertion in this scope
    # GLOBAL | NONLOCAL - not LOCAL
    # NONLOCAL | LOCAL - can it happen?
    # GLOBAL | LOCAL - can it happen?
    UNKNOWN = GLOBAL | NONLOCAL | LOCAL


# TODO: rename to Symbol and look at variables that use this class
@attr.s(auto_attribs=True)
class SymbolData:
    # it seems like symbol's scope doesn't change
    # throughout the scope execution
    scope: ScopeType = ScopeType.UNKNOWN

    imported: bool = False  # TODO: handle this in acr translation
    is_arg: bool = False
    holds_symbol_table: bool = False

    type: PynalyserType = UnknownType

    # if we change from UNKNOWN to more specific it's fine
    # but if we change from specific to other specific than
    # probably something went wrong
    # is it the responsibility of this class to check this?
    def change_scope(self, new_scope: ScopeType, fail: bool = True) -> bool:
        # TODO: enable more options:
        # from unspecific to specific
        # GLOBAL | NONLOCAL, NONLOCAL | LOCAL, GLOBAL | LOCAL
        # is ok to go to
        # GLOBAL or NONLOCAL, NONLOCAL or LOCAL, GLOBAL or LOCAL
        if self.scope is ScopeType.UNKNOWN:
            self.scope = new_scope
            return True
        elif self.scope == new_scope:
            return True
        else:
            if not fail:
                return False

            # TODO: change to custom class
            raise ValueError(
                f"changing the `scope` from {self.scope} to {new_scope}")


class SymbolTable(DefaultDict[str, SymbolData]):
    def __init__(self, *args, **kwargs):
        super().__init__(SymbolData, *args, **kwargs)



import ast


class AstNormalizer(ast.NodeTransformer):
    def visit_Num(self, node: ast.Num) -> ast.Constant:
        return ast.Constant(node.n)

    def visit_Str(self, node: ast.Str) -> ast.Constant:
        return ast.Constant(node.s)

    def visit_Bytes(self, node: ast.Bytes) -> ast.Constant:
        return ast.Constant(node.s)

    def visit_NameConstant(self, node: ast.NameConstant) -> ast.Constant:
        return ast.Constant(node.value)

    def visit_Ellipsis(self, node: ast.Ellipsis) -> ast.Constant:
        return ast.Constant(...)

    def visit_Index(self, node: ast.Index) -> ast.AST:
        return self.visit(node.value)

    def visit_ExtSlice(self, node: ast.ExtSlice) -> ast.Tuple:
        return self.visit(ast.Tuple(node.dims, ast.Load()))

    # We are just getting rid of deprecated nodes,
    # so we don't need a warning (see: NodeTransformer.visit_Constant)
    # Same as NodeTransformer.visit(node) with visit_Constant removed
    def visit_Constant(self, node: ast.Constant) -> ast.Constant:
        return self.generic_visit(node)  # type: ignore


def normalize_ast(tree: ast.AST) -> ast.AST:
    return AstNormalizer().visit(tree)


def normalize_ast_module(module: ast.Module) -> ast.Module:
    tree = normalize_ast(module)
    assert isinstance(tree, ast.Module)
    return tree



import sys
from typing import Any, List, Union

from pynalyser.acr.utils import NODE

from .. import portable_ast as ast
from ..acr import classes as acr_c
from .symbols import ScopeType, SymbolTable
from .tools import Analyser
from .types import Arg, Arguments, FunctionType, SymTabType


class AssignVisitor(ast.NodeVisitor):
    # accounted for:
    # Attribute - don't visit fields, XXX: for now
    # Subscript - don't visit fields
    # Starred - just visit itself
    # Name - add name, if we got here
    # List - just visit itself
    # Tuple - just visit itself
    # other nodes AFAIK will not appear here

    names: List[str] = []

    def get_names(self, node: ast.AST) -> List[str]:
        self.names.clear()
        self.visit(node)
        return self.names

    def visit_Attribute(self, node: ast.Attribute) -> None:
        pass

    def visit_Subscript(self, node: ast.Subscript) -> None:
        pass

    def visit_Name(self, node: ast.Name) -> None:
        self.names.append(node.id)


# TODO: scope should be separated from acr
# it should yield it's oun result
class ScopeAnalyser(Analyser):
    # we receive acr with global and nonlocal keywords
    # already transformed into symbol's scope value
    # but everything else is unknown and have to be filled here

    assign_visitor: AssignVisitor = AssignVisitor()

    def handle_scope(self, scope: acr_c.Scope) -> None:
        if scope.is_symbol:
            # this symbol is used in self.scope
            self.scope.symbol_table[scope.name]

    def handle_arg(self, scope: acr_c.Scope, name: str) -> Arg:
        symbol = scope.symbol_table[name]
        symbol.is_arg = True

        if not symbol.change_scope(ScopeType.LOCAL, fail=False):
            raise SyntaxError(
                f"duplicate argument '{name}' in function definition")

        return Arg(name, symbol)

    def handle_function(
        self, scope: Union[acr_c.Lambda, acr_c.Function]
    ) -> None:

        assert len(scope.symbol_table) == 0

        args = Arguments()
        symbol = self.scope.symbol_table[scope.name]
        symbol.type = FunctionType(args)
        symbol.change_scope(ScopeType.LOCAL)
        symbol.holds_symbol_table = True
        scope.symbol_table = symbol.type.symbol_table

        if sys.version_info >= (3, 8):
            for arg in scope.args.posonlyargs:
                args.posargs.append(self.handle_arg(scope, arg.arg))

        for arg in scope.args.args:
            args.args.append(self.handle_arg(scope, arg.arg))

        for arg in scope.args.kwonlyargs:
            args.kwargs.append(self.handle_arg(scope, arg.arg))

        if scope.args.vararg is not None:
            args.stararg = self.handle_arg(scope, scope.args.vararg.arg)

        if scope.args.kwarg is not None:
            args.twostararg = self.handle_arg(scope, scope.args.kwarg.arg)

    def visit_For(self, node: acr_c.For) -> None:
        self.setup_symbols_by_assign(node.target)

    # def visit_ListComp(self, node: acr_c.ListComp) -> None:
    #     self.handle_scope(node)

    # def visit_SetComp(self, node: acr_c.SetComp) -> None:
    #     self.handle_scope(node)

    # def visit_GeneratorExp(self, node: acr_c.GeneratorExp) -> None:
    #     self.handle_scope(node)

    # def visit_DictComp(self, node: acr_c.DictComp) -> None:
    #     self.handle_scope(node)

    def visit_Lambda(self, node: acr_c.Lambda) -> None:
        self.handle_function(node)

    def visit_Function(self, node: acr_c.Function) -> None:
        self.handle_function(node)

    def visit_Class(self, node: acr_c.Class) -> None:
        self.handle_scope(node)

    def setup_symbols_by_assign(self, *targets: ast.AST) -> None:
        names = []
        for sub_node in targets:
            names.extend(self.assign_visitor.get_names(sub_node))

        for name in names:
            symbol_data = self.scope.symbol_table[name]

            # in the other case it's already defined
            symbol_data.change_scope(ScopeType.LOCAL)
            symbol_data.imported = False

    def visit_Assign(self, node: ast.Assign) -> None:
        self.setup_symbols_by_assign(*node.targets)

    def visit_AugAssign(self, node: ast.AugAssign) -> None:
        pass

    def visit_AnnAssign(self, node: ast.AnnAssign) -> None:
        # XXX: should be only one name in the end:
        self.setup_symbols_by_assign(node.target)

    def setup_symbols_by_import(self, targets: List[ast.alias]) -> None:
        for alias in targets:
            name = alias.asname or alias.name  # those are never == ""
            symbol_data = self.scope.symbol_table[name]
            symbol_data.imported = True

    def visit_Import(self, node: ast.Import) -> None:
        self.setup_symbols_by_import(node.names)

    def visit_ImportFrom(self, node: ast.ImportFrom) -> None:
        self.setup_symbols_by_import(node.names)

    def visit_NamedExpr(self, node: ast.NamedExpr) -> None:
        names = self.assign_visitor.get_names(node.target)
        assert len(names) == 1
        name, = names

        symbol = self.scope.symbol_table[name]
        if isinstance(self.scope, acr_c.Comprehension):
            # in the other case it's already defined
            if not symbol.change_scope(ScopeType.NONLOCAL, fail=False):
                raise NotImplementedError(
                    "NamedExpr should make symbol local for enclosing scope")
        else:
            # in the other case it's already defined
            symbol.change_scope(ScopeType.LOCAL)

        symbol.imported = False

    def visit_Name(self, node: ast.Name) -> None:
        # this name have been used in this scope
        self.scope.symbol_table[node.id]



from collections import defaultdict
from typing import DefaultDict, Dict, List, Optional, Tuple, TypeVar, Union

import attr

from .. import portable_ast as ast
from ..analysis._types import PynalyserType, UnknownType
from ..analysis.symbols import SymbolTable

ACR_T = TypeVar("ACR_T")


@attr.s
class ACR:
    """The base class for each class of abstract representation of the code.
    Inherited classes could be dumpable by `acr.dump`,
    but for that to work properly inherited class should use `attr.s`.
    """

    _attributes: Tuple[str, ...] = attr.ib(init=False, default=())
    _fields: Tuple[str, ...] = attr.ib(init=False, default=())
    # _nonblock_fields: Tuple[str, ...] = ()
    # _block_fields: Tuple[str, ...] = attr.ib(init=False, default=())

    # XXX: abc ? from_ast, to_ast
    # @classmethod
    # def from_ast(cls: Type[ACR_T], node: ast.AST) -> ACR_T:
    #     raise NotImplementedError

    # FIXME: this should run only on class creation, so metaclasses?
    def __attrs_post_init__(self):
        _fields = list(attr.fields_dict(type(self)).keys())

        for name in self._attributes:
            _fields.remove(name)
        _fields.remove("_attributes")
        _fields.remove("_fields")

        self._fields = tuple(_fields)

        # _nonblock_fields = []
        # _block_fields = []  currently in Block


@attr.s(auto_attribs=True)
class Name(ACR):
    name: str
    is_symbol: bool = attr.ib(init=False, default=False)

    _attributes: Tuple[str, ...] = attr.ib(init=False, default=("is_symbol",))


@attr.s(auto_attribs=True)
class Asyncable(ACR):
    is_async: bool = attr.ib(kw_only=True)


CODE = Union[
    # stmt - from scope body
    ast.Delete, ast.Assign, ast.AugAssign, ast.AnnAssign,
    ast.Import, ast.ImportFrom,
    # Nonlocal, Global - we have symbol's scope for that
    # Expr - no need for it
    ast.Pass,  # so blocks only with pass will be fine
    ast.Break, ast.Continue,

    # expr - from breakdown of complex expressions
    ast.BoolOp, ast.NamedExpr, ast.BinOp, ast.UnaryOp,
    ast.IfExp, ast.Dict, ast.Set, ast.Await,
    ast.Yield, ast.YieldFrom,  # special ones, can be removed later
    ast.Compare, ast.Call, ast.JoinedStr,
    # FormattedValue should not exist by itself (without JoinedStr)
    ast.Constant, ast.Attribute, ast.Subscript,
    # Starred should not exist by itself (without Call or Assign)
    ast.Name,  # the only use for that to exit by itself
    # is reporting a NameError: name 'xxx' is not defined
    ast.List, ast.Tuple,
    # Slice can appear only in Subscript

    # not ast
    "ScopeReference"
]


CONTROL_FLOW = Union[  # TODO: not finished?
    "CodeBlock", "Block",

    # stmt - from scope body
    ast.Return, ast.Raise, ast.Assert,
    ast.Break, ast.Continue,

    # expr - from breakdown of complex expressions
    # ast.Yield, ast.YieldFrom for now it's in the CODE
]


# XXX: ControlFlowSomething? BlockContainer?
class FlowContainer(ACR, List[CONTROL_FLOW]):
    def get_code_block(self) -> "CodeBlock":
        if len(self):
            block = self[-1]
            if type(block) is not CodeBlock:
                block = CodeBlock()
                self.append(block)
        else:
            block = CodeBlock()
            self.append(block)

        return block

    def add_code(self, code: CODE) -> None:
        """Add code to latest block, if it's is not `CodeBlock`
        function will add new `CodeBlock` to `blocks` and add code there"""

        self.get_code_block().append(code)


class CodeBlock(ACR, List[CODE]):
    """a.k.a. Basic block"""
    pass


# @attr.s(auto_attribs=True)
class Block(ACR):
    _block_fields: Tuple[str, ...] = ()

    def __attrs_post_init__(self):
        super().__attrs_post_init__()
        _fields = list(self._fields)
        _fields.remove("_block_fields")
        self._fields = tuple(_fields)


@attr.s(auto_attribs=True)
class BodyBlock(Block):
    body: FlowContainer = attr.ib(factory=FlowContainer, init=False)
    _block_fields: Tuple[str, ...] = attr.ib(init=False, default=("body",))


@attr.s(auto_attribs=True)
class ScopeReference(ACR, ast.AST):  # ast so NodeTransformer could handle this
    """Refers to a single scope in the `.scopes`
    of the local scope.

    It replaces actual definition in the and gets put in the `CODE`
    instead of scope.
    """
    name: str
    id: int

    def get_scope(self, local_scope: "Scope") -> "Scope":
        return local_scope.scopes[self.name][self.id]


# XXX: do we need this class?
class ScopeDefs(Dict[int, "Scope"]):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(self, *args, **kwargs)
        self._last_index = 0

    def append(self, value: "Scope") -> None:
        self[self._last_index] = value
        self._last_index += 1

    @property
    def last_index(self) -> int:
        return self._last_index


@attr.s(auto_attribs=True)
class Scope(Name, BodyBlock):
    scopes: DefaultDict[str, ScopeDefs] = attr.ib(
        init=False, factory=lambda: defaultdict(ScopeDefs))
    symbol_table: SymbolTable = attr.ib(init=False, factory=SymbolTable)
    # parent? probably nay

    def add_scope(self, scope: "Scope") -> ScopeReference:
        defs = self.scopes[scope.name]
        reference = ScopeReference(scope.name, defs.last_index)
        defs.append(scope)
        return reference


class Module(Scope):
    """`name` is the name of the file that this module belongs to
    """
    # path?
    # XXX: is_symbol = True?
    pass


@attr.s(auto_attribs=True)
class ACRWithAttributes(ACR):
    lineno: int = attr.ib(kw_only=True)
    col_offset: int = attr.ib(kw_only=True)
    # TODO: make end_lineno and end_col_offset required fields
    # after switching to python-version-independent ast generator
    # (it will generate lastest ast for earlsier python versions)
    end_lineno: Optional[int] = attr.ib(default=None, kw_only=True)
    end_col_offset: Optional[int] = attr.ib(default=None, kw_only=True)

    _attributes: Tuple[str, ...] = attr.ib(init=False, default=(
        "lineno", "col_offset", "end_lineno", "end_col_offset"))


@attr.s(auto_attribs=True)
class ScopeWithAttributes(ACRWithAttributes, Scope):
    _attributes: Tuple[str, ...] = attr.ib(init=False, default=(
        "lineno", "col_offset", "end_lineno", "end_col_offset", "is_symbol"))


# we don't care about 'type_ignores'

@attr.s(auto_attribs=True)
class Class(ScopeWithAttributes):
    bases: List[ast.expr]  # parent-classes
    keywords: List[ast.keyword]  # initially metaclass and kws for it
    decorator_list: List[ast.expr]
    metaclass: Optional[type] = None  # XXX: it's probably not type really

    is_symbol: bool = attr.ib(init=False, default=True)


@attr.s(auto_attribs=True)
class Function(ScopeWithAttributes, Asyncable):
    # TODO: create my out typed attributes (+ think about ease of str dumping)
    args: ast.arguments = attr.ib(factory=ast.arguments)
    decorator_list: List[ast.expr] = attr.ib(factory=list)

    return_type: PynalyserType = attr.ib(init=False, default=UnknownType)
    is_symbol: bool = attr.ib(init=False, default=True)


LAMBDA_NAME = "<lambda>"


@attr.s(auto_attribs=True)
class Lambda(ScopeWithAttributes):
    name: str = attr.ib(default=LAMBDA_NAME, init=False)
    args: ast.arguments = attr.ib(factory=ast.arguments)


@attr.s(auto_attribs=True)
class Comprehension(ScopeWithAttributes):
    # XXX: shouldn't have `body` from Scope,
    # but removing this will bring currently unneeded refactoring
    generators: List[ast.comprehension] = attr.ib(kw_only=True)
    _block_fields: Tuple[str, ...] = attr.ib(init=False, default=())


@attr.s(auto_attribs=True)
class EltComprehension(Comprehension):
    elt: ast.expr


@attr.s(auto_attribs=True)
class ListComp(EltComprehension):
    name: str = attr.ib(init=False, default="<listcomp>")


@attr.s(auto_attribs=True)
class SetComp(EltComprehension):
    name: str = attr.ib(init=False, default="<setcomp>")


@attr.s(auto_attribs=True)
class GeneratorExp(EltComprehension):
    name: str = attr.ib(init=False, default="<genexpr>")


@attr.s(auto_attribs=True)
class DictComp(Comprehension):
    key: ast.expr
    value: ast.expr
    name: str = attr.ib(init=False, default="<dictcomp>")


@attr.s(auto_attribs=True)
class MatchCase(BodyBlock):
    pattern: ast.pattern
    guard: Optional[ast.expr]


@attr.s(auto_attribs=True)
class Match(ACRWithAttributes, Block):
    subject: ast.expr
    cases: List[MatchCase] = attr.ib(init=False, factory=list)
    _block_fields: Tuple[str, ...] = attr.ib(init=False, default=("cases",))


@attr.s(auto_attribs=True)
class With(ACRWithAttributes, BodyBlock, Asyncable):
    items: List[ast.withitem]


@attr.s(auto_attribs=True)
class BodyElseBlock(BodyBlock):
    orelse: FlowContainer = attr.ib(factory=FlowContainer, init=False)
    _block_fields: Tuple[str, ...] = attr.ib(
        init=False, default=("body", "orelse"))


@attr.s(auto_attribs=True)
class If(ACRWithAttributes, BodyElseBlock):
    test: ast.expr


@attr.s(auto_attribs=True)
class ExceptHandler(ACRWithAttributes, BodyBlock):
    type: Optional[ast.expr]
    name: Optional[str]


@attr.s(auto_attribs=True)
class Try(ACRWithAttributes, BodyElseBlock):
    handlers: List[ExceptHandler] = attr.ib(factory=list, init=False)
    finalbody: FlowContainer = attr.ib(factory=FlowContainer, init=False)
    _block_fields: Tuple[str, ...] = attr.ib(
        init=False, default=("body", "handlers", "orelse", "finalbody"))


@attr.s(auto_attribs=True)
class Loop(ACRWithAttributes, BodyElseBlock):  # XXX: do we need this class?
    pass


@attr.s(auto_attribs=True)
class For(Loop, Asyncable):
    target: ast.expr  # ? Union[ast.Name, ast.Tuple, ast.List]
    iter: ast.expr


@attr.s(auto_attribs=True)
class While(Loop):
    test: ast.expr


# idea:
# class "list"
# def mark(index) -> id
# marked item's index will be tracked
# def get_index(id) -> index



import ast
from typing import Any, List, Tuple, Union
from warnings import warn

from ..acr import classes as acr_c
from ..acr.utils import NODE, NodeVisitor, dump
from ._ref_types import (BinOpType, CallType, ItemType, SubscriptType,
                         SymbolType, CompareType)
from ._types import (DUNDER_SIGNATURE, UnknownType, AnyType, IntType, ListType,
                     PynalyserType, SequenceType, SingleType, SliceType,
                     TupleType, UnionType)
from ._symb_types import FunctionType, Arguments, Arg, SymTabType
from .tools import Analyser


# XXX: merge into TypeInference?
class ExprTypeInference(NodeVisitor):
    # auto_generic_visit: bool = True

    def infer(self, scope: acr_c.Scope, node: NODE) -> PynalyserType:
        res = self.start(scope, node)
        if isinstance(res, PynalyserType):
            return res
        return AnyType

    def visit_Call(self, node: ast.Call) -> PynalyserType:
        return CallType(
            self.visit(node.func),
            tuple(self.visit(item) for item in node.args),
            tuple((item.arg, self.visit(item.value)) for item in node.keywords)
        )

    # def visit_Attribute(self, node: ast.Attribute) -> PynalyserType:
    #     return AnyType

    def visit_Subscript(self, node: ast.Subscript) -> PynalyserType:
        return SubscriptType(self.visit(node.value), self.visit(node.slice))

    def visit_BinOp(self, node: ast.BinOp) -> PynalyserType:
        return BinOpType(type(node.op).__name__,
                         self.visit(node.left), self.visit(node.right))

    def visit_Compare(self, node: ast.Compare) -> PynalyserType:
        return CompareType(self.visit(node.left),
                           [type(op).__name__ for op in node.ops],
                           [self.visit(item) for item in node.comparators])

    ### Comprehentions ###

    def visit_ListComp(self, node: acr_c.ListComp) -> PynalyserType:
        return ListType(item_type=UnknownType)  # TODO: infer item type

    ### Builtin sequences ###

    def visit_List(self, node: ast.List) -> PynalyserType:
        return ListType(item_type=UnionType.make(
            *map(self.visit, node.elts), fallback=UnknownType))

    def visit_Tuple(self, node: ast.Tuple) -> PynalyserType:
        return TupleType(item_type=UnionType.make(
            *map(self.visit, node.elts), fallback=UnknownType))

    ### Basic "building blocks"  ###

    def visit_Slice(self, node: ast.Slice) -> PynalyserType:
        return SliceType()

    def visit_Constant(self, node: ast.Constant) -> PynalyserType:
        if isinstance(node.value, int):
            return IntType()
        return SingleType(name=type(node.value).__name__, is_builtin=True)

    def visit_Name(self, node: ast.Name) -> PynalyserType:
        return SymbolType(node.id, self.scope.symbol_table[node.id])


class TypeInference(Analyser):
    expr_type_inference: ExprTypeInference = ExprTypeInference()

    def infer_acr_expr(self, node: Union[ast.AST, acr_c.ACR]) -> PynalyserType:
        return self.expr_type_inference.infer(self.scope, node)

    def infer_assignment(self, node: ast.AST, tp: PynalyserType) -> None:
        if isinstance(node, ast.Name):
            symbol_data = self.scope.symbol_table[node.id]
            if symbol_data.type is UnknownType:
                symbol_data.type = tp
            else:
                symbol_data.type = UnionType.make(symbol_data.type, tp)
                # XXX: ideally we have to create
                # new variables-clones for type changes
                # anyways it's not handled here
                # raise TypeError("symbol's type is already set to {}")

        # in case of list or tuple we can infer number of elements
        # elif isinstance(node, ast.Name):

    def visit_Assign(self, node: ast.Assign) -> None:
        tp = self.infer_acr_expr(node.value)
        for target in node.targets:
            self.infer_assignment(target, tp)

    def visit_AnnAssign(self, node: ast.AnnAssign) -> None:
        if node.value is not None:
            tp = self.infer_acr_expr(node.value)
            self.infer_assignment(node.target, tp)

    def visit_For(self, node: acr_c.For) -> None:
        tp = self.infer_acr_expr(node.iter)
        self.infer_assignment(node.target, ItemType(tp))

    def visit_While(self, node: acr_c.While) -> None:
        tp = self.infer_acr_expr(node.test)
        # can be boolable
        # if bool(test) is always True / False,
        # then we have infinite loop or
        # we can skip loop



import sys
from typing import List, NoReturn, Optional, Union

from .. import portable_ast as ast
from ..analysis.symbols import ScopeType
from .classes import (Block, Class, DictComp, ExceptHandler, FlowContainer,
                      For, Function, GeneratorExp, If, Lambda, ListComp, Match,
                      MatchCase, Module, Scope, ScopeReference, SetComp, Try,
                      While, With)

STMT_SCOPE = Union[ast.FunctionDef, ast.AsyncFunctionDef, ast.ClassDef]


class ForBlocks(ast.AST):
    _fields = "blocks",
    blocks: FlowContainer


# XXX: restrict visit, so if there's no visit_x then raise exception
class Translator(ast.NodeTransformer):
    scope: Scope
    container: FlowContainer

    #### Transformations used only for expr scopes ####

    #### Blocks (all stmt) ####

    # XXX: we can shrink down ast.AST to smaller subset
    def handle_block_without_appending(
            self, block: Block, node: ast.AST) -> None:

        self.generic_visit(block)  # type: ignore

        self.handle_fields_of_block(block, node)

    def handle_fields_of_block(self, block: Block, node: ast.AST) -> None:
        prev_container = self.container

        for name in block._block_fields:
            value = getattr(node, name)
            some_container = getattr(block, name)  # FIXME: bad name!
            if isinstance(some_container, FlowContainer):
                self.container = some_container
                self.generic_visit(ForBlocks(value[:]))
                setattr(block, name, self.container)
            elif isinstance(some_container, list):
                # container should never be needed here
                for sub_container in value:
                    func = getattr(
                        self, "my_visit_" + type(sub_container).__name__)
                    some_container.append(func(sub_container))
            else:
                raise TypeError("All `_block_fields` should be subclasses "
                                "of the `list` or `FlowContainer`")

        self.container = prev_container

    def handle_block(self, block: Block, node: ast.AST) -> None:
        self.container.append(block)
        self.handle_block_without_appending(block, node)

    def visit_match_case(self, node: ast.match_case) -> ast.match_case:
        self.handle_block(
            MatchCase(node.pattern, node.guard),  # no attributes
            node)
        return node

    def visit_Match(self, node: ast.Match) -> ast.Match:
        self.handle_block(
            Match(node.subject,
                  lineno=node.lineno, col_offset=node.col_offset),
            node)
        return node

    def visit_With(self, node: ast.With) -> ast.With:
        self.handle_block(
            With(node.items, is_async=False,
                 lineno=node.lineno, col_offset=node.col_offset),
            node)
        return node

    def visit_AsyncWith(self, node: ast.AsyncWith) -> ast.AsyncWith:
        self.handle_block(
            With(node.items, is_async=True,
                 lineno=node.lineno, col_offset=node.col_offset),
            node)
        return node

    def visit_If(self, node: ast.If) -> ast.If:
        self.handle_block(
            If(node.test,
               lineno=node.lineno, col_offset=node.col_offset),
            node)
        return node

    def my_visit_ExceptHandler(self, node: ast.ExceptHandler) -> ExceptHandler:
        # should only be called from handle_block_without_appending
        acr_handler = ExceptHandler(
            node.type, node.name,
            lineno=node.lineno, col_offset=node.col_offset)
        self.handle_block_without_appending(acr_handler, node)
        return acr_handler

    def visit_Try(self, node: ast.Try) -> ast.Try:
        block = Try(lineno=node.lineno, col_offset=node.col_offset)
        self.handle_block(block, node)
        return node

    def visit_For(self, node: ast.For) -> ast.For:
        self.handle_block(
            For(node.target, node.iter, is_async=False,
                lineno=node.lineno, col_offset=node.col_offset),
            node)
        return node

    def visit_AsyncFor(self, node: ast.AsyncFor) -> ast.AsyncFor:
        self.handle_block(
            For(node.target, node.iter, is_async=True,
                lineno=node.lineno, col_offset=node.col_offset),
            node)
        return node

    def visit_While(self, node: ast.While) -> ast.While:
        self.handle_block(
            While(node.test,
                  lineno=node.lineno, col_offset=node.col_offset),
            node)
        return node

    #### expr scopes ####

    # XXX: we can shrink down node's ast.AST to smaller subset
    def handle_scope(self, scope: Scope, node: ast.AST) -> ScopeReference:
        reference = self.scope.add_scope(scope)

        self.generic_visit(scope)  # type: ignore

        prev_scope, self.scope = self.scope, scope
        self.handle_fields_of_block(scope, node)
        self.scope = prev_scope

        return reference

    def visit_Lambda(self, node: ast.Lambda) -> ScopeReference:
        lamb = Lambda(
            node.args, lineno=node.lineno, col_offset=node.col_offset)
        node.body = [ast.Expr(node.body)]  # type: ignore
        return self.handle_scope(lamb, node)

    def visit_ListComp(self, node: ast.ListComp) -> ScopeReference:
        return self.handle_scope(
            ListComp(node.elt, generators=node.generators,
                     lineno=node.lineno, col_offset=node.col_offset),
            node)

    def visit_SetComp(self, node: ast.SetComp) -> ScopeReference:
        return self.handle_scope(
            SetComp(node.elt, generators=node.generators,
                    lineno=node.lineno, col_offset=node.col_offset),
            node)

    def visit_GeneratorExp(self, node: ast.GeneratorExp) -> ScopeReference:
        return self.handle_scope(
            GeneratorExp(node.elt, generators=node.generators,
                         lineno=node.lineno, col_offset=node.col_offset),
            node)

    def visit_DictComp(self, node: ast.DictComp) -> ScopeReference:
        return self.handle_scope(
            DictComp(node.key, node.value, generators=node.generators,
                     lineno=node.lineno, col_offset=node.col_offset),
            node)

    #### stmt scopes ####

    def translate_from_module(self, module: ast.Module,
                              name: str) -> Module:
        acr = self.scope = Module(name)
        self.container = acr.body
        self.generic_visit(module)
        del self.scope, self.container
        return acr

    def visit_Module(self, node: ast.Module) -> NoReturn:
        raise NotImplementedError(  # XXX: parsing error idk
            "Either you just called `visit` on the tree "
            "(which starts with `ast.Module`) - use `translate_from_module` "
            "or there's more than one `ast.Module` in the tree")

    def handle_stmt_scope(self, scope: Scope, node: STMT_SCOPE) -> None:
        self.container.add_code(self.handle_scope(scope, node))

    def visit_FunctionDef(self, node: ast.FunctionDef) -> ast.FunctionDef:
        func = Function(
            node.name, node.args, node.decorator_list, is_async=False,
            lineno=node.lineno, col_offset=node.col_offset)
        self.handle_stmt_scope(func, node)
        return node

    def visit_AsyncFunctionDef(
        self, node: ast.AsyncFunctionDef
    ) -> ast.AsyncFunctionDef:

        func = Function(
            node.name, node.args, node.decorator_list, is_async=True,
            lineno=node.lineno, col_offset=node.col_offset)
        self.handle_stmt_scope(func, node)
        return node

    def visit_ClassDef(self, node: ast.ClassDef) -> ast.ClassDef:
        self.handle_stmt_scope(
            Class(
                node.name, node.bases, node.keywords, node.decorator_list,
                lineno=node.lineno, col_offset=node.col_offset),
            node)
        return node

    #### Other contlow flow ####

    def visit_Return(self, node: ast.Return) -> ast.Return:
        self.generic_visit(node)
        self.container.append(node)
        return node

    def visit_Raise(self, node: ast.Raise) -> ast.Raise:
        self.generic_visit(node)
        self.container.append(node)
        return node

    def visit_Assert(self, node: ast.Assert) -> ast.Assert:
        self.generic_visit(node)
        self.container.append(node)
        return node

    def visit_Break(self, node: ast.Break) -> ast.Break:
        self.generic_visit(node)
        self.container.append(node)
        return node

    def visit_Continue(self, node: ast.Continue) -> ast.Continue:
        self.generic_visit(node)
        self.container.append(node)
        return node

    # TODO: handle yield
    # def visit_Yield(self, node: ast.Yield) -> ast.Yield:
    #     self.generic_visit(node)
    #     self.container.append(node)
    #     return node

    # TODO: handle yield from
    # def visit_YieldFrom(self, node: ast.YieldFrom) -> ast.YieldFrom:
    #     self.generic_visit(node)
    #     self.container.append(node)
    #     return node

    #### Leftover stmt-s ####

    def visit_Delete(self, node: ast.Delete) -> ast.Delete:
        self.generic_visit(node)
        self.container.add_code(node)
        return node

    def visit_Assign(self, node: ast.Assign) -> ast.Assign:
        self.generic_visit(node)
        self.container.add_code(node)
        return node

    def visit_AugAssign(self, node: ast.AugAssign) -> ast.AugAssign:
        self.generic_visit(node)
        self.container.add_code(node)
        return node

    def visit_AnnAssign(self, node: ast.AnnAssign) -> ast.AnnAssign:
        self.generic_visit(node)
        self.container.add_code(node)
        return node

    def visit_Import(self, node: ast.Import) -> ast.Import:
        self.generic_visit(node)
        self.container.add_code(node)
        return node

    def visit_ImportFrom(self, node: ast.ImportFrom) -> ast.ImportFrom:
        self.generic_visit(node)
        self.container.add_code(node)
        return node

    def visit_Global(self, node: ast.Global) -> ast.Global:
        # no need to visit - Global(names: list[str])
        for name in node.names:
            symbol_data = self.scope.symbol_table[name]
            # generally imports before global should not be allowed,
            # but cpython allows it https://tiny.one/global-in-docs
            symbol_data.change_scope(ScopeType.GLOBAL)
        return node

    def visit_Nonlocal(self, node: ast.Nonlocal) -> ast.Nonlocal:
        # no need to visit - Nonlocal(names: list[str])
        for name in node.names:
            symbol_data = self.scope.symbol_table[name]
            # generally imports before nonlocal should not be allowed,
            # but cpython allows it https://tiny.one/nonlocal-in-docs
            symbol_data.change_scope(ScopeType.NONLOCAL)
        return node

    def visit_Expr(self, node: ast.Expr) -> ast.Expr:
        self.generic_visit(node)
        self.container.add_code(node.value)  # type: ignore
        return node

    def visit_Pass(self, node: ast.Pass) -> ast.Pass:
        self.generic_visit(node)
        self.container.add_code(node)
        return node

    #### Leftover expr-s ####

    # BoolOp

    def visit_NamedExpr(self, node: ast.NamedExpr) -> ast.NamedExpr:
        self.generic_visit(node)
        self.container.add_code(node)
        return node

    # BinOp
    # UnaryOp

    # IfExp
    # Dict
    # Set

    # Await
    # Yield
    # YieldFrom

    # Compare
    # Call
    # FormattedValue
    # JoinedStr
    # Constant

    # Attribute
    # Subscript
    # Starred

    # Name

    # List
    # Tuple

    # Slice


def translate_ast_to_acr(tree: ast.Module, name: str) -> Module:
    return Translator().translate_from_module(tree, name)



from typing import List, Optional
from ._types import PynalyserType, UnknownType
from .symbols import SymbolData, SymbolTable

import attr


class SymTabType(PynalyserType):
    symbol_table: SymbolTable = attr.ib(init=False, factory=SymbolTable)


@attr.s(auto_attribs=True, hash=True)
class Arg:
    name: str
    symbol: SymbolData
    default: Optional[PynalyserType] = None


@attr.s(auto_attribs=True, hash=True)
class Arguments:
    posargs: List[Arg] = attr.ib(factory=list)
    args: List[Arg] = attr.ib(factory=list)
    stararg: Optional[Arg] = None
    kwargs: List[Arg] = attr.ib(factory=list)
    twostararg: Optional[Arg] = None


@attr.s(auto_attribs=True, hash=True)
class FunctionType(SymTabType):
    args: Arguments
    return_type: PynalyserType = UnknownType



import os
from glob import glob

from setuptools import find_packages, setup

from pynalyser import __name__, __version__

requirements = {}
for path in glob("requirements/*.txt"):
    with open(path) as file:
        name = os.path.basename(path)[:-4]
        requirements[name] = [line.strip() for line in file]

with open("README.md") as file:
    long_description = file.read()

github_link = "https://github.com/0dminnimda/{0}".format(__name__)

setup(
    name=__name__,
    version=__version__,
    description="Static Python Code Analyzer",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="0dminnimda",
    author_email="0dminnimda.contact@gmail.com",
    url=github_link,
    packages=find_packages(),
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Software Development :: Pre-processors",
        "Topic :: Utilities",
        "Typing :: Typed",
    ],
    license="MIT",
    project_urls={
        "Bug tracker": github_link + "/issues",
    },
    install_requires=requirements.pop("basic"),
    python_requires=">=3.6",
    extras_require=requirements,
    package_data={__name__: ["py.typed"]},
)



import ast
from collections import defaultdict
from typing import (Any, Collection, List, NamedTuple, Optional, Tuple,
                    TypeVar, Union)

from .classes import (ACR, Block, CodeBlock, FlowContainer, Module, Scope,
                      ScopeReference)

# Dumping


class Context(NamedTuple):
    annotate_fields: bool
    include_attributes: bool
    indent: Optional[str]


def dump(
        obj: Union[ACR, ast.AST], annotate_fields: bool = True,
        include_attributes: bool = False, *,
        indent: Optional[Union[str, int]] = None) -> str:
    if not isinstance(obj, (ACR, ast.AST, list, dict)):
        raise TypeError(  # XXX: should we force it?
            f"expected one of the AST / ACR / list / dict, "
            "got {type(obj).__name__}")
    if indent is not None and not isinstance(indent, str):
        indent = " " * indent
    return _format(
        obj, Context(annotate_fields, include_attributes, indent), lvl=0)[0]


def _format(obj: Any, ctx: Context, lvl: int) -> Tuple[str, bool]:
    if ctx.indent is not None:
        lvl += 1
        prefix = "\n" + ctx.indent * lvl
        sep = ",\n" + ctx.indent * lvl
    else:
        prefix = ""
        sep = ", "

    if isinstance(obj, list):
        args = []
        allsimple = True

        for item in obj:
            value, simple = _format(item, ctx, lvl)
            allsimple = allsimple and simple
            args.append(value)

        value, allsimple = _format_args_with_allsimple(
            prefix, sep, args, allsimple)

        value = f"[{value}]"
        if type(obj) is not list:
            value = f"{type(obj).__name__}({value})"
        return value, allsimple

    if isinstance(obj, dict):
        args = []
        allsimple = True

        for key, value in obj.items():
            # dict key should be "simple"
            value, simple = _format(value, ctx, lvl)
            allsimple = allsimple and simple
            args.append(f"{key!r}: {value}")

        value, allsimple = _format_args_with_allsimple(
            prefix, sep, args, allsimple)

        value = f"{{{value}}}"
        if isinstance(obj, defaultdict):
            value = f"{type(obj).__name__}({obj.default_factory}, {value})"
        elif type(obj) is not dict:
            value = f"{type(obj).__name__}({value})"

        return value, allsimple

    if isinstance(obj, (ACR, ast.AST)):
        value, allsimple = _format_args_with_allsimple(
            prefix, sep, *_format_ast_or_acr(obj, ctx, lvl))
        return f"{type(obj).__name__}({value})", allsimple

    return repr(obj), True


def _format_args_with_allsimple(prefix: str, sep: str, args: Collection[str],
                                allsimple: bool) -> Tuple[str, bool]:
    if allsimple and len(args) <= 3:
        return ", ".join(args), not args
    return prefix + sep.join(args), False


def _format_attr(inst: Any, name: str, ctx: Context,
                 lvl: int) -> Tuple[str, bool]:
    return _format(
        getattr(inst, name, "<Is not an attribute of the object>"), ctx, lvl)


def _format_ast_or_acr(obj: Union[ast.AST, ACR], ctx: Context,
                       lvl: int) -> Tuple[List[str], bool]:
    args = []
    allsimple = True

    for name in obj._fields:
        value, simple = _format_attr(obj, name, ctx, lvl)
        allsimple = allsimple and simple
        if ctx.annotate_fields:
            args.append(f"{name}={value}")
        else:
            args.append(value)

    if ctx.include_attributes and obj._attributes:
        for name in obj._attributes:
            value, simple = _format_attr(obj, name, ctx, lvl)
            allsimple = allsimple and simple
            args.append(f"{name}={value}")

    return args, allsimple


# Tree traversing


NODE = Union[ACR, ast.AST]


class NodeVisitor:
    _ast_visitor = ast.NodeVisitor

    scope: Scope
    block: Block

    strict: bool = False
    auto_generic_visit: bool = True

    def start(self, init_scope_block: Scope,
              to_visit: Optional[NODE] = None) -> Any:

        self.scope = self.block = init_scope_block

        if to_visit is None:
            result = self.visit(init_scope_block)
        else:
            result = self.visit(to_visit)

        del self.scope, self.block
        return result

    def acr_generic_visit(self, node: NODE) -> Any:
        # handle acr
        if isinstance(node, Scope):
            previous_scope = self.scope
            self.scope = node

            previous_block = self.block
            self.block = node

            try:
                return self.generic_visit(node)
            finally:
                self.scope = previous_scope
                self.block = previous_block

        if isinstance(node, Block):
            previous_block = self.block
            self.block = node

            try:
                return self.generic_visit(node)
            finally:
                self.block = previous_block

        return self.generic_visit(node)

    def visit(self, node: NODE) -> Any:
        method = 'visit_' + type(node).__name__
        visitor = getattr(self, method, None)

        if visitor is None:
            if isinstance(node, ScopeReference):
                return self.generic_visit(node)
            if self.strict:
                raise ValueError(
                    f"There are no '{method}' method. "
                    "You see this message because you're in strict mode. "
                    f"See {type(self).__name__}.strict")

            result = self.acr_generic_visit(node)
        else:
            result = visitor(node)
            if self.auto_generic_visit:
                self.acr_generic_visit(node)

        return result

    def generic_visit(self, node: NODE) -> Any:
        if isinstance(node, ScopeReference):
            return self.visit(node.get_scope(self.scope))

        if isinstance(node, ast.AST):
            return self._ast_visitor.generic_visit(self, node)  # type: ignore

        if isinstance(node, Block):
            for name in node._block_fields:
                self.visit(getattr(node, name))
            return node

        if isinstance(node, FlowContainer):
            for item in node:
                self.visit(item)
            return node

        if isinstance(node, CodeBlock):
            for code in node:
                self.visit(code)
            return node

        raise RuntimeError(
            f"Expected ACR or AST, but got {type(node).__name__}")


class ACRCodeTransformer(NodeVisitor):
    """Allows to change contents of the code (of the CodeBlock)"""

    _ast_visitor = ast.NodeTransformer

    def generic_visit(self, node: NODE) -> Any:
        if isinstance(node, ScopeReference):
            return self.visit(node.get_scope(self.scope))

        if isinstance(node, ast.AST):
            return self._ast_visitor.generic_visit(self, node)  # type: ignore

        if isinstance(node, Block):
            for name in node._block_fields:
                self.visit(getattr(node, name))
            return node

        if isinstance(node, FlowContainer):
            for i, item in enumerate(node):
                node[i] = self.visit(item)
            return node

        if isinstance(node, CodeBlock):
            new_code_block: CodeBlock = CodeBlock()
            for code in node:
                value = self.visit(code)
                if value is None:
                    pass
                elif isinstance(value, (ast.AST, ACR)):
                    new_code_block.append(value)  # type: ignore
                else:
                    new_code_block.extend(value)
            return node

        raise RuntimeError(
            f"Expected ACR or AST, but got {type(node).__name__}")



import ast
import io
import os
from typing import List, Optional

from .acr.classes import Module
from .acr.translation import translate_ast_to_acr
from .analysis.types import TypeInference
from .analysis.scope import ScopeAnalyser
from .analysis.tools import Analyser, AnalysisContext
from .normalize_ast import normalize_ast_module


def parse(path: str) -> Module:
    return parse_open(path, "r", encoding="utf-8")


def parse_open(*args, **kwargs) -> Module:
    with open(*args, **kwargs) as stream:
        assert isinstance(stream, io.TextIOBase)
        return parse_stream(stream)  # type: ignore


def parse_stream(stream: io.TextIOWrapper) -> Module:
    with stream:
        return parse_ast(
            os.path.splitext(os.path.basename(stream.name))[0],
            ast.parse(stream.read()))


def parse_ast(name: str, module: ast.Module) -> Module:
    return translate_ast_to_acr(normalize_ast_module(module), name)


def analyse_context(analysers: List[Analyser], ctx: AnalysisContext) -> None:
    for analyser in analysers:
        analyser.analyse(ctx)


DEFAULT_ANALYSERS = [
    ScopeAnalyser(),
    TypeInference(),
]


def analyse(ctx: AnalysisContext,
            analysers: Optional[List[Analyser]] = None) -> AnalysisContext:

    if analysers is None:
        analysers = DEFAULT_ANALYSERS
    assert analysers is not None  # for typechecker

    analyse_context(analysers, ctx)

    return ctx


def analyse_file(
    source: str, analysers: Optional[List[Analyser]] = None
) -> AnalysisContext:

    ctx = AnalysisContext([parse(source)])

    return analyse(ctx, analysers)



from typing import Any, Dict, List

import attr

from ..acr import classes as acr_c
from ..acr.utils import NodeVisitor


@attr.s(auto_attribs=True)
class AnalysisContext:
    modules: List[acr_c.Module]
    results: Dict[str, Any] = attr.ib(init=False, factory=lambda: dict())


class Analyser(NodeVisitor):
    context: AnalysisContext

    def analyse(self, ctx: AnalysisContext) -> None:
        self.start(ctx.modules[0])
        self.context = ctx
